/*
 * W E X T R A C T
 *
 *  Extract a specified time interval from an audio file.
 *
 *  Usage: wextract -i inputfile -o outputfile -s StartTime -e EndTime
 *
 *  Mark Liberman <myl@cis.upenn.edu> 11/30/2006
 *
 *    Tested on Linux (Redhat Enterprise 4) and Ubuntu (Edgy Eft)
 *      and on Windows (Cygwin under XP)
 *
 * Copyright (C) 2006  Mark Y. Liberman
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_SNDFILE_H
#include <sndfile.h>
#endif

#ifdef HAVE_LOCALE_H
#include <locale.h>
#endif

static char pgname[] = "wextract";
static char version[]=PACKAGE_VERSION;
static char compdate[]=__DATE__ " " __TIME__ ;

void
Version(void){
  fprintf(stderr,"%s %s \n",pgname,version);
  fprintf(stderr,"Compiled %s.\n",compdate);
  fprintf(stderr,"Copyright 2006 Mark Y. Liberman\n");
  fprintf(stderr,"Released under the terms of the GNU General Public License.\n");
  fprintf(stderr,"Report bugs to: myl@ldc.upenn.edu.\n");
  putc('\n',stderr);
}

void
Usage(void){
  fprintf(stderr,"Extract a segment from an audio file.\n");
  fprintf(stderr,"Usage:\t%s -i <input file>\n\t\t -o <output file>\n\t\t -s <start time>\n\t\t -e <end time>\n",
	  pgname);
  fprintf(stderr,"\t%s -h \t\t print this help message\n",pgname);
  fprintf(stderr,"\t%s -v \t\t print version information\n",pgname);
#ifndef HAVE_LIBSNDFILE
  fprintf(stderr,"Only simple WAVE files containing linear PCM data are supported.\n");
#else
  fprintf(stderr,
	  "The file types and audio formats supported are those supported by libsndfile.\n");
#endif
  putc('\n',stderr);
}

#ifndef HAVE_LIBSNDFILE
typedef struct {
  /* RiffChunk */
  unsigned char riffID[4]; /* 'RIFF' */
  long RchunkSize;  /* filesize - 8 */
  unsigned char waveID[4]; /* 'WAVE' */

  /* FormatChunk */
  unsigned char FchunkID[4]; /* 'fmt ' */
  long FchunkSize;
  short wFormatTag;
  unsigned short wChannels;
  unsigned long dwSamplesPerSec;
  unsigned long dwAvgBytesPerSec;
  unsigned short wBlockAlign;
  unsigned short wBitsPerSample;

  /* Note: there may be additional fields here, if wFormatTag is not = 1 */
  /* DataChunk  -- we assume only one */
  unsigned char DchunkID[4]; /* 'data' */
  long DchunkSize; /* filesize - 44 */
} WavHeader;

WavHeader InHeader, OutHeader;
#endif

double StrToSeconds (char *s) {
  double sec;
  char *cptr;

  cptr = strchr(s,':');
  if(!cptr) sec = atof(s);
  else {
    sec = atof(cptr+1);
    *cptr='\0';
    sec += 60 * atoi(s);
  }
  return sec;
}

main(int ac, char **av)
{
  double StartTime = 0.0, EndTime = 0.0, FileSeconds;
  long int firstbyte, lastbyte, totalbytes, filesize, ret;
  char *InFileName=NULL, *OutFileName=NULL;
  int c, nread;
  unsigned long FrameOffset;	/* Frame at which to begin extraction */
  unsigned long Frames;		/* Number of frames to extract */
  double Seconds;		/* Duration of segment to extract */

#ifndef HAVE_LIBSNDFILE
  unsigned char *data;
  FILE *infd, *outfd;
#else
  SNDFILE *infd;		/* Input file descriptor */
  SNDFILE *outfd;		/* Output file descriptor */
  struct SF_INFO InputInfo;
  long SamplesRead;	
  long SamplesWritten;	
  double *sbuf;
#endif
  extern char *optarg;
  extern int optopt;

#ifdef HAVE_SETLOCALE
   setlocale(LC_ALL,"");
#endif

  while((c=getopt(ac,av,":he:i:o:s:v"))!=EOF)
    switch(c){
    case 'i':
      InFileName = optarg;
      break;
    case 'o':
      OutFileName = optarg;
      break;
    case 's':
      StartTime = StrToSeconds(optarg);
      break;
    case 'e':
      EndTime =  StrToSeconds(optarg);
      break;
    case 'h':
      Usage();
      exit(2);
      break;			/* NOTREACHED */
    case 'v':
      Version();
      exit(2);
      break;			/* NOTREACHED */
    case ':':
       fprintf(stderr,"%s: missing argument to option flag %c.\n",pgname,optopt);
       exit(3);
    default:
      fprintf(stderr,"%s: illegal option flag %c\n",pgname,optopt);
      exit(4);
    }

   if(ac < 9) {
     Usage();
     exit(2);
   }


  if(!InFileName || !OutFileName) {
    fprintf(stderr,"%s: the names of the input and output files must be specified.\n",pgname);
    exit(7);
  }

  if(StartTime < 0) {
    fprintf(stderr,"The start time must be greater than 0.\n");
    exit(3);
  }
  Seconds = EndTime - StartTime;
  if(Seconds < 0) {
    fprintf(stderr,"The end time must be greater than the start time.\n");
    exit(3);
  }
  if (Seconds == 0.0) {
    fprintf(stderr,"There is no point extracting a region 0 seconds long.\n");
    exit(5);
  }

#ifndef HAVE_LIBSNDFILE
  infd = fopen(InFileName, "r");
  if(infd==NULL){
    fprintf(stderr,"Can't open %s for reading\n", InFileName);
    exit(2);
  }
  outfd = fopen(OutFileName, "w");
  if(outfd==NULL){
    fprintf(stderr,"Can't open %s for writing\n", OutFileName);
    exit(2);
  }
  nread = fread(&InHeader, 44, 1, infd);
  if(nread != 1){
    fprintf(stderr, "Failed to read header from %s\n", InFileName);
    exit(2);
  }
  if(StartTime == 0.0 && EndTime == 0.0){
    Usage();
    exit(2);
  }

  OutHeader = InHeader;

  FileSeconds = (double)(InHeader.DchunkSize)/(double)InHeader.dwAvgBytesPerSec;
  if (EndTime == 0.0) EndTime = FileSeconds;

  if(EndTime > FileSeconds){
    fprintf(stderr, 
	    "The end time of %g seconds exceeds the length of the file,\nwhich is %g.\n",
	    EndTime,FileSeconds);
    exit(2);
  }
  firstbyte = 44 + (int)(StartTime*(InHeader.dwAvgBytesPerSec));
  totalbytes = (int)((EndTime-StartTime)*(InHeader.dwAvgBytesPerSec));
  totalbytes += totalbytes%(InHeader.wBlockAlign);
  Frames = (EndTime - StartTime) * InHeader.dwSamplesPerSec;

  if(fseek(infd, firstbyte, SEEK_SET)){
    fprintf(stderr, "Couldn't seek to byte %d in %s\n", firstbyte,InFileName);
    exit(2);
  }

  OutHeader.RchunkSize = totalbytes + 44 - 8;
  OutHeader.DchunkSize = totalbytes;

  ret = fwrite(&OutHeader, 44, 1, outfd);
  data = malloc(totalbytes);
  if(data == NULL) {
    fprintf(stderr,"%s: unable to allocate storage for audio data.\n",pgname);
    fclose(infd); fclose(outfd); unlink(outfd);
    exit(5);
  }
  ret = fread(data, 1, totalbytes, infd);
  ret = fwrite(data, 1, totalbytes, outfd);

  fclose(infd); fclose(outfd);

#else
  /* Open files */
  InputInfo.format = 0;
  infd = sf_open(InFileName,SFM_READ,&InputInfo);
  if(infd == NULL){
    fprintf(stderr,"%s: could not open file %s for read.\n",pgname,InFileName);
    exit(2);
  }

  FileSeconds = (double) InputInfo.frames / (double) InputInfo.samplerate;

  if(EndTime > FileSeconds){
    fprintf(stderr, 
	    "%s: the end time of %g seconds exceeds the length of the file,\nwhich is %g.\n",
	    pgname,EndTime,FileSeconds);
    exit(2);
  }

  Frames = (unsigned long) (Seconds * (double) InputInfo.samplerate);
  FrameOffset = (unsigned long) (StartTime * (double)  InputInfo.samplerate);

  outfd = sf_open(OutFileName,SFM_WRITE,&InputInfo);
  if (outfd == NULL){
    fprintf(stderr,"%s: could not open file %s for write.\n",pgname,OutFileName);
    exit(2);
  }

  sbuf = (double *) malloc(Frames * sizeof(double));
  if(sbuf == NULL) {
    fprintf(stderr,"%s: unable to allocate storage for audio data.\n",pgname);
    sf_close(infd); sf_close(outfd); unlink(outfd);
    exit(5);
  }

  if (sf_seek(infd,FrameOffset,SEEK_SET) == -1) {
    fprintf(stderr,"%s: seek failed.\n",pgname);
    exit(6);
  }
  SamplesRead = sf_read_double(infd,sbuf,Frames);
  if(sf_error(infd)){
    fprintf(stderr,"%s: error reading file %s:\n",pgname,InFileName);
    fprintf(stderr,"\t%s\n",sf_strerror(infd));
    exit(4);
  }

  SamplesWritten = sf_write_double(outfd,sbuf,SamplesRead);
  if( (SamplesWritten != SamplesRead) || (sf_error(outfd))){
    fprintf(stderr,"%s: error writing file %s:\n",pgname,OutFileName);
    fprintf(stderr,"\t%s\n",sf_strerror(outfd));
    sf_close(infd); sf_close(outfd); unlink(outfd);
    exit(4);
  }
  
  sf_close(infd);
  sf_close(outfd);

#endif

#ifdef HAVE_PRINTF_THSEP
  fprintf(stderr,"Region [%4.3f,%4.3f] (%'lu samples) of %s written to %s.\n",
	  StartTime,EndTime,Frames,InFileName,OutFileName);
#else
  fprintf(stderr,"Region [%4.3f,%4.3f] (%lu samples) of %s written to %s.\n",
	  StartTime,EndTime,Frames,InFileName,OutFileName);
#endif
}
